@@ -12,7 +12,9 @@ class Loggable:
     """
     Base class that provides consistently templated logging.
 
-    Inspired by wasabi's good/info/warn/fail methods.
+    Inspired by `wasabi`'s `good`/`info`/`warn`/`fail` methods.
+
+    [`Rich` style guide](https://rich.readthedocs.io/en/latest/style.html)
     """
 
     CONSOLE = Console()
@@ -22,7 +24,8 @@ class Loggable:
 
     def _good(self, message):
         self.__class__.CONSOLE.print(
-            f":green_circle: {self.__class__.__name__}: {message}", style="green"
+            f":green_circle: {self.__class__.__name__}: {message}",
+            style="green",
         )
 
     def _info(self, message):
@@ -32,7 +35,8 @@ class Loggable:
 
     def _warn(self, message):
         self.__class__.CONSOLE.print(
-            f":yellow_circle: {self.__class__.__name__}: {message}", style="yellow"
+            f":yellow_circle: {self.__class__.__name__}: {message}",
+            style="yellow",
         )
 
     def _fail(self, message):
