@@ -1,26 +1,41 @@
 """
-Dataset objects which extend beyond DataFrames.
-Specifically, we need a collection of DataFrames where rows can be transferred cleanly and columns can be transformed easily.
+???+ note "Dataset classes which extend beyond DataFrames."
+
+    When we supervise a collection of data, these operations need to be simple:
+
+    -   managing `raw`/`train`/`dev`/`test` subsets
+    -   transferring data points between subsets
+    -   pulling updates from annotation interfaces
+    -   pushing updates to annotation interfaces
+    -   getting a 2D embedding
+    -   loading data for training models
 """
 import pandas as pd
 import numpy as np
 from tqdm import tqdm
 from hover import module_config
 from hover.core import Loggable
-from bokeh.models import Button, Dropdown, ColumnDataSource, DataTable, TableColumn
-from .local_config import dataset_help_widget
+from hover.utils.bokeh_helper import auto_label_color
+from bokeh.models import (
+    Button,
+    Dropdown,
+    ColumnDataSource,
+    DataTable,
+    TableColumn,
+    HTMLTemplateFormatter,
+)
+from .local_config import dataset_help_widget, COLOR_GLYPH_TEMPLATE
 
 
 class SupervisableDataset(Loggable):
     """
-    Feature-agnostic class for a dataset open to supervision.
-    Raw -- piecewise annoatation -> Gold -> Dev/Test
-    Raw -- batch annotation -> Noisy -> Train
+    ???+ note "Feature-agnostic class for a dataset open to supervision."
 
-    Keeping a DataFrame form and a list-of-dicts ("dictl") form, with the intention that
-    - the DataFrame form supports most kinds of operations;
-    - the list-of-dicts form could be useful for manipulations outside the scope of pandas;
-    - synchronization between the two forms should be called sparingly.
+        Keeping a DataFrame form and a list-of-dicts ("dictl") form, with the intention that
+
+        - the DataFrame form supports most kinds of operations;
+        - the list-of-dicts form could be useful for manipulations outside the scope of pandas;
+        - synchronization between the two forms should be called sparingly.
     """
 
     # 'scratch': intended to be directly editable by other objects, i.e. Explorers
@@ -45,14 +60,15 @@ class SupervisableDataset(Loggable):
         label_key="label",
     ):
         """
-        Initialize the dataset with dictl and df forms; initialize the mapping between categorical-int and string labels.
-
-        - param raw_dictl: a list of dicts holding the raw data that DO NOT have annotation.
-        - param train_dictl: a list of dicts holding the batch-annotated noisy train set.
-        - param dev_dictl: a list of dicts holding the gold dev set.
-        - param test_dictl: a list of dicts holding the gold test set.
-        - param feature_key: key in each piece of dict mapping to the feature.
-        - param label_key: key in each piece of dict mapping to the ground truth in STRING form.
+        ???+ note "Create (1) dictl and df forms and (2) the mapping between categorical and string labels."
+            | Param         | Type   | Description                          |
+            | :------------ | :----- | :----------------------------------- |
+            | `raw_dictl`   | `list` | list of dicts holding the **to-be-supervised** raw data |
+            | `train_dictl` | `list` | list of dicts holding any **supervised** train data |
+            | `dev_dictl`   | `list` | list of dicts holding any **supervised** dev data   |
+            | `test_dictl`  | `list` | list of dicts holding any **supervised** test data  |
+            | `feature_key` | `str`  | the key for the feature in each piece of data |
+            | `label_key`   | `str`  | the key for the `**str**` label in supervised data |
         """
         self._info("Initializing...")
 
@@ -102,7 +118,10 @@ class SupervisableDataset(Loggable):
 
     def copy(self, use_df=True):
         """
-        Create another instance, carrying over the data entries.
+        ???+ note "Create another instance, copying over the data entries."
+            | Param    | Type   | Description                          |
+            | :------- | :----- | :----------------------------------- |
+            | `use_df` | `bool` | whether to use the df or dictl form |
         """
         if use_df:
             self.synchronize_df_to_dictl()
@@ -117,7 +136,7 @@ class SupervisableDataset(Loggable):
 
     def setup_widgets(self):
         """
-        Critical widgets for interactive data management.
+        ???+ note "Create `bokeh` widgets for interactive data management."
         """
         self.update_pusher = Button(
             label="Push", button_type="success", height_policy="fit", width_policy="min"
@@ -172,7 +191,7 @@ class SupervisableDataset(Loggable):
 
     def view(self):
         """
-        Defines the layout of bokeh models.
+        ???+ note "Defines the layout of `bokeh` objects when visualized."
         """
         # local import to avoid naming confusion/conflicts
         from bokeh.layouts import row, column
@@ -185,9 +204,13 @@ class SupervisableDataset(Loggable):
 
     def subscribe_update_push(self, explorer, subset_mapping):
         """
-        Enable pushing updated DataFrames to explorers that depend on them.
+        ???+ note "Enable pushing updated DataFrames to explorers that depend on them."
+            | Param            | Type   | Description                            |
+            | :--------------- | :----- | :------------------------------------- |
+            | `explorer`       | `BokehBaseExplorer` | the explorer to register  |
+            | `subset_mapping` | `dict` | `dataset` -> `explorer` subset mapping |
 
-        Note: the reason we need this is due to `self.dfs[key] = ...`-like assignments. If DF operations were all in-place, then the explorers could directly access the updates through their `self.dfs` references.
+            Note: the reason we need this is due to `self.dfs[key] = ...`-like assignments. If DF operations were all in-place, then the explorers could directly access the updates through their `self.dfs` references.
         """
         # local import to avoid import cycles
         from hover.core.explorer.base import BokehBaseExplorer
@@ -206,7 +229,11 @@ class SupervisableDataset(Loggable):
 
     def subscribe_data_commit(self, explorer, subset_mapping):
         """
-        Enable committing data across subsets, specified by a selection in an explorer and a dropdown widget of the dataset.
+        ???+ note "Enable committing data across subsets, specified by a selection in an explorer and a dropdown widget of the dataset."
+            | Param            | Type   | Description                            |
+            | :--------------- | :----- | :------------------------------------- |
+            | `explorer`       | `BokehBaseExplorer` | the explorer to register  |
+            | `subset_mapping` | `dict` | `dataset` -> `explorer` subset mapping |
         """
 
         def callback_commit(event):
@@ -242,6 +269,8 @@ class SupervisableDataset(Loggable):
                 self._info(
                     f"Committed {valid_slice.shape[0]} (valid out of {sel_slice.shape[0]} selected) entries from {sub_k} to {sub_to} ({size_before} -> {size_after} with {size_mid-size_after} overwrites)."
                 )
+            # chain another callback
+            self._callback_update_population()
 
         self.data_committer.on_click(callback_commit)
         self._good(
@@ -250,8 +279,13 @@ class SupervisableDataset(Loggable):
 
     def setup_label_coding(self, verbose=True, debug=False):
         """
-        Auto-determine labels in the dataset, then create encoder/decoder in lexical order.
-        Add ABSTAIN as a no-label placeholder.
+        ???+ note "Auto-determine labels in the dataset, then create encoder/decoder in lexical order."
+            Add `"ABSTAIN"` as a no-label placeholder which gets ignored categorically.
+
+            | Param     | Type   | Description                        |
+            | :-------- | :----- | :--------------------------------- |
+            | `verbose` | `bool` | whether to log verbosely           |
+            | `debug`   | `bool` | whether to enable label validation |
         """
         all_labels = set()
         for _key in [*self.__class__.PUBLIC_SUBSETS, *self.__class__.PRIVATE_SUBSETS]:
@@ -277,7 +311,11 @@ class SupervisableDataset(Loggable):
 
     def validate_labels(self, raise_exception=True):
         """
-        Check that every label is in the encoder.
+        ???+ note "Assert that every label is in the encoder."
+
+            | Param             | Type   | Description                         |
+            | :---------------- | :----- | :---------------------------------- |
+            | `raise_exception` | `bool` | whether to raise errors when failed |
         """
         for _key in [*self.__class__.PUBLIC_SUBSETS, *self.__class__.PRIVATE_SUBSETS]:
             _invalid_indices = None
@@ -291,6 +329,13 @@ class SupervisableDataset(Loggable):
                     raise ValueError("invalid labels")
 
     def setup_pop_table(self, **kwargs):
+        """
+        ???+ note "Set up a bokeh `DataTable` widget for monitoring subset data populations."
+
+            | Param      | Type   | Description                  |
+            | :--------- | :----- | :--------------------------- |
+            | `**kwargs` |        | forwarded to the `DataTable` |
+        """
         subsets = [
             *self.__class__.SCRATCH_SUBSETS,
             *self.__class__.PUBLIC_SUBSETS,
@@ -303,6 +348,11 @@ class SupervisableDataset(Loggable):
                 TableColumn(field=f"count_{_subset}", title=_subset)
                 for _subset in subsets
             ],
+            TableColumn(
+                field="color",
+                title="color",
+                formatter=HTMLTemplateFormatter(template=COLOR_GLYPH_TEMPLATE),
+            ),
         ]
         self.pop_table = DataTable(source=pop_source, columns=pop_columns, **kwargs)
 
@@ -315,7 +365,10 @@ class SupervisableDataset(Loggable):
 
             # re-compute label population
             eff_labels = [module_config.ABSTAIN_DECODED, *self.classes]
-            pop_data = dict(label=eff_labels)
+            color_dict = auto_label_color(self.classes)
+            eff_colors = [color_dict[_label] for _label in eff_labels]
+
+            pop_data = dict(color=eff_colors, label=eff_labels)
             for _subset in subsets:
                 _subpop = self.dfs[_subset]["label"].value_counts()
                 pop_data[f"count_{_subset}"] = [
@@ -326,16 +379,18 @@ class SupervisableDataset(Loggable):
             pop_source.data = pop_data
 
             self._good(
-                f"Pop updater: latest population with {len(self.classes)} classes"
+                f"Pop updater: latest population with {len(self.classes)} classes."
             )
 
         update_population()
-        self.data_committer.on_click(update_population)
         self.dedup_trigger.on_click(update_population)
 
+        # store the callback so that it can be referenced by other methods
+        self._callback_update_population = update_population
+
     def df_deduplicate(self):
         """
-        Cross-deduplicate data entries by feature between subsets.
+        ???+ note "Cross-deduplicate data entries by feature between subsets."
         """
         self._info("Deduplicating...")
         # for data entry accounting
@@ -374,7 +429,7 @@ class SupervisableDataset(Loggable):
 
     def synchronize_dictl_to_df(self):
         """
-        Re-make dataframes from lists of dictionaries.
+        ???+ note "Re-make dataframes from lists of dictionaries."
         """
         self.dfs = dict()
         for _key, _dictl in self.dictls.items():
@@ -389,7 +444,7 @@ class SupervisableDataset(Loggable):
 
     def synchronize_df_to_dictl(self):
         """
-        Re-make lists of dictionaries from dataframes.
+        ???+ note "Re-make lists of dictionaries from dataframes."
         """
         self.dictls = dict()
         for _key, _df in self.dfs.items():
@@ -397,7 +452,14 @@ class SupervisableDataset(Loggable):
 
     def compute_2d_embedding(self, vectorizer, method, **kwargs):
         """
-        Get embeddings in the xy-plane and return the reducer.
+        ???+ note "Get embeddings in the xy-plane and return the dimensionality reducer."
+            Reference: [`DimensionalityReducer`](https://github.com/phurwicz/hover/blob/main/hover/core/representation/reduction.py)
+
+            | Param        | Type       | Description                        |
+            | :----------- | :--------- | :--------------------------------- |
+            | `vectorizer` | `callable` | the feature -> vector function     |
+            | `method`     | `str`      | arg for `DimensionalityReducer`    |
+            | `**kwargs`   |            | kwargs for `DimensionalityReducer` |
         """
         from hover.core.representation.reduction import DimensionalityReducer
 
@@ -454,12 +516,13 @@ class SupervisableDataset(Loggable):
 
     def loader(self, key, vectorizer, batch_size=64, smoothing_coeff=0.0):
         """
-        Prepare a Torch Dataloader for training or evaluation.
-
-        - param key(str): the subset of dataset to use.
-        - param vectorizer(callable): callable that turns a string into a vector.
-        - param smoothing_coeff: the smoothing coeffient for soft labels.
-          - type smoothing_coeff: float
+        ???+ note "Prepare a torch `Dataloader` for training or evaluation."
+            | Param        | Type       | Description                        |
+            | :----------- | :--------- | :--------------------------------- |
+            | `key`        | `str`      | subset of data, e.g. `"train"`     |
+            | `vectorizer` | `callable` | the feature -> vector function     |
+            | `batch_size` | `int`      | size per batch                     |
+            | `smoothing_coeff` | `float` | portion of probability to equally split between classes |
         """
         # lazy import: missing torch should not break the rest of the class
         from hover.utils.torch_helper import vector_dataloader, one_hot, label_smoothing
@@ -492,7 +555,7 @@ class SupervisableDataset(Loggable):
 
 class SupervisableTextDataset(SupervisableDataset):
     """
-    Can add text-specific methods.
+    ???+ note "Can add text-specific methods."
     """
 
     FEATURE_KEY = "text"
