@@ -1,21 +1,22 @@
-"""Intermediate classes based on the functionality."""
+"""
+???+ note "Intermediate classes based on the functionality."
+"""
 import numpy as np
 from bokeh.models import CDSView, IndexFilter
 from bokeh.palettes import Category20
-from bokeh.transform import factor_cmap
 from hover import module_config
 from hover.utils.misc import current_time
-from .local_config import bokeh_hover_tooltip
+from .local_config import bokeh_hover_tooltip, SOURCE_COLOR_FIELD, SOURCE_ALPHA_FIELD
 from .base import BokehBaseExplorer
 
 
 class BokehDataFinder(BokehBaseExplorer):
     """
-    Plot data points in grey ('gainsboro') and highlight search positives in coral.
+    ???+ note "Plot data points in grey ('gainsboro') and highlight search positives in coral."
 
-    Features:
+        Features:
 
-    - the search widgets will highlight the results through a change of color, which is arguably the best visual effect.
+        -   the search widgets will highlight the results through a change of color, which gives arguably the best focus.
     """
 
     SUBSET_GLYPH_KWARGS = {
@@ -31,7 +32,13 @@ class BokehDataFinder(BokehBaseExplorer):
     }
 
     def plot(self, *args, **kwargs):
-        """Plot the data map."""
+        """
+        ???+ note "Plot all data points."
+            | Param      | Type  | Description           |
+            | :--------- | :---- | :-------------------- |
+            | `*args`    |       | for consistency       |
+            | `**kwargs` |       | for consistency       |
+        """
         for _key, _source in self.sources.items():
             self.figure.circle(
                 "x", "y", name=_key, source=_source, **self.glyph_kwargs[_key]
@@ -41,12 +48,11 @@ class BokehDataFinder(BokehBaseExplorer):
 
 class BokehDataAnnotator(BokehBaseExplorer):
     """
-    Annoate data points via callbacks.
+    ???+ note "Annoate data points via callbacks on the buttons."
 
-    Features:
+        Features:
 
-    - alter values in the 'label' column through the widgets.
-    - **SERVER ONLY**: only works in a setting that allows Python callbacks.
+        - alter values in the 'label' column through the widgets.
     """
 
     SUBSET_GLYPH_KWARGS = {
@@ -60,9 +66,23 @@ class BokehDataAnnotator(BokehBaseExplorer):
         for _key in ["raw", "train", "dev", "test"]
     }
 
+    def _postprocess_sources(self):
+        """
+        ???+ note "Infer glyph colors from the label dynamically."
+        """
+        # infer glyph color from labels
+        color_dict = self.auto_color_mapping()
+
+        def get_color(label):
+            return color_dict.get(label, "gainsboro")
+
+        for _key, _df in self.dfs.items():
+            _color = _df["label"].apply(get_color).tolist()
+            self.sources[_key].add(_color, SOURCE_COLOR_FIELD)
+
     def _setup_widgets(self):
         """
-        Create annotator widgets and assign Python callbacks.
+        ???+ note "Create annotator widgets and assign Python callbacks."
         """
         from bokeh.models import TextInput, Button, Dropdown
 
@@ -100,11 +120,10 @@ class BokehDataAnnotator(BokehBaseExplorer):
             self.dfs["raw"].at[selected_idx, "label"] = label
             example_new = self.dfs["raw"].at[selected_idx[0], "label"]
             self._good(
-                f"Applied {len(selected_idx)} annotations: {label} (e.g. {example_old} -> {example_new}"
+                f"Applied {len(selected_idx)} annotations: {label} (e.g. {example_old} -> {example_new})"
             )
 
             self._update_sources()
-            self.plot()
             self._good(f"Updated annotator plot at {current_time()}")
 
         def callback_export(event, path_root=None):
@@ -150,23 +169,18 @@ class BokehDataAnnotator(BokehBaseExplorer):
         self.annotator_apply.on_click(self._callback_subset_display)
         self.annotator_export.on_click(self._callback_export)
 
-    @BokehBaseExplorer.auto_legend
     def plot(self):
         """
-        Re-plot with the new labels.
-
-        Overrides the parent method.
-        Determines the label->color mapping dynamically.
+        ???+ note "Re-plot all data points with the new labels."
+            Overrides the parent method.
+            Determines the label -> color mapping dynamically.
         """
-        labels, palette = self.auto_labels_palette()
-
         for _key, _source in self.sources.items():
             self.figure.circle(
                 "x",
                 "y",
                 name=_key,
-                color=factor_cmap("label", palette=palette, factors=labels),
-                legend_group="label",
+                color=SOURCE_COLOR_FIELD,
                 source=_source,
                 **self.glyph_kwargs[_key],
             )
@@ -175,13 +189,13 @@ class BokehDataAnnotator(BokehBaseExplorer):
 
 class BokehSoftLabelExplorer(BokehBaseExplorer):
     """
-    Plot data points according to their labels and confidence scores.
+    ???+ note "Plot data points according to their labels and confidence scores."
 
-    Features:
+        Features:
 
-    - the predicted label will correspond to fill_color.
-    - the confidence score, assumed to be a float between 0.0 and 1.0, will be reflected through fill_alpha.
-    - currently not considering multi-label scenarios.
+        - the predicted label will correspond to fill_color.
+        - the confidence score, assumed to be a float between 0.0 and 1.0, will be reflected through fill_alpha.
+        - currently not considering multi-label scenarios.
     """
 
     SUBSET_GLYPH_KWARGS = {
@@ -191,10 +205,18 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
 
     def __init__(self, df_dict, label_col, score_col, **kwargs):
         """
-        On top of the requirements of the parent class,
-        the input dataframe should contain:
+        ???+ note "Additional construtor"
+            On top of the requirements of the parent class,
+            the input dataframe should contain:
 
-        (1) label_col and score_col for "soft predictions".
+            - label_col and score_col for "soft predictions".
+
+            | Param       | Type   | Description                  |
+            | :---------- | :----- | :--------------------------- |
+            | `df_dict`   | `dict` | `str` -> `DataFrame` mapping |
+            | `label_col` | `str`  | column for the soft label    |
+            | `score_col` | `str`  | column for the soft score    |
+            | `**kwargs`  |        | forwarded to `bokeh.plotting.figure` |
         """
         assert label_col != "label", "'label' field is reserved"
         self.label_col = label_col
@@ -202,14 +224,22 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
         super().__init__(df_dict, **kwargs)
 
     def _build_tooltip(self):
-        """Extending from the parent method."""
+        """
+        ???+ note "On top of the parent method, add the soft label fields to the tooltip."
+        """
         return bokeh_hover_tooltip(
             **self.__class__.TOOLTIP_KWARGS,
             custom={"Soft Label": self.label_col, "Soft Score": self.score_col},
         )
 
     def _setup_dfs(self, df_dict, **kwargs):
-        """Extending from the parent method."""
+        """
+        ???+ note "On top of the parent method, add filler values to additional columns."
+            | Param      | Type   | Description                  |
+            | :--------- | :----- | :--------------------------- |
+            | `df_dict`  | `dict` | `str` -> `DataFrame` mapping |
+            | `**kwargs` |        | forwarded to the parent method |
+        """
         super()._setup_dfs(df_dict, **kwargs)
 
         for _key, _df in self.dfs.items():
@@ -218,19 +248,50 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
             if self.score_col not in _df.columns:
                 _df[self.score_col] = 0.5
 
-    @BokehBaseExplorer.auto_legend
+    def _postprocess_sources(self):
+        """
+        ???+ note "Infer glyph colors from the label dynamically."
+        """
+        # infer glyph color from labels
+        color_dict = self.auto_color_mapping()
+
+        def get_color(label):
+            return color_dict.get(label, "gainsboro")
+
+        # infer glyph alpha from pseudo-percentile of soft label scores
+        scores = np.concatenate(
+            [_df[self.score_col].tolist() for _df in self.dfs.values()]
+        )
+        scores_mean = scores.mean()
+        scores_std = scores.std() + 1e-4
+
+        def pseudo_percentile(confidence, lower=0.1, upper=0.9):
+            # pretend that 2*std on each side covers everything
+            unit_shift = upper - lower / 4
+            # shift = unit_shift * z_score
+            shift = unit_shift * (confidence - scores_mean) / scores_std
+            percentile = 0.5 + shift
+            return min(upper, max(lower, percentile))
+
+        # infer alpha from score percentiles
+        for _key, _df in self.dfs.items():
+            _color = _df[self.label_col].apply(get_color).tolist()
+            _alpha = _df[self.score_col].apply(pseudo_percentile).tolist()
+            self.sources[_key].add(_color, SOURCE_COLOR_FIELD)
+            self.sources[_key].add(_alpha, SOURCE_ALPHA_FIELD)
+
     def plot(self, **kwargs):
         """
-        Plot the confidence map.
+        ???+ note "Plot all data points, setting color alpha based on the soft score."
+            | Param      | Type   | Description                  |
+            | :--------- | :----- | :--------------------------- |
+            | `**kwargs` |        | forwarded to plotting markers |
         """
-        labels, palette = self.auto_labels_palette()
-
         for _key, _source in self.sources.items():
             # prepare plot settings
             preset_kwargs = {
-                "legend_group": self.label_col,
-                "color": factor_cmap(self.label_col, palette=palette, factors=labels),
-                "fill_alpha": self.score_col,
+                "color": SOURCE_COLOR_FIELD,
+                "fill_alpha": SOURCE_ALPHA_FIELD,
             }
             eff_kwargs = self.glyph_kwargs[_key].copy()
             eff_kwargs.update(preset_kwargs)
@@ -242,13 +303,13 @@ class BokehSoftLabelExplorer(BokehBaseExplorer):
 
 class BokehMarginExplorer(BokehBaseExplorer):
     """
-    Plot data points along with two versions of labels.
-    Could be useful for A/B tests.
+    ???+ note "Plot data points along with two versions of labels."
+        Could be useful for A/B tests.
 
-    Features:
+        Features:
 
-    - can choose to only plot the margins about specific labels.
-    - currently not considering multi-label scenarios.
+        - can choose to only plot the margins about specific labels.
+        - currently not considering multi-label scenarios.
     """
 
     SUBSET_GLYPH_KWARGS = {
@@ -261,17 +322,31 @@ class BokehMarginExplorer(BokehBaseExplorer):
 
     def __init__(self, df_dict, label_col_a, label_col_b, **kwargs):
         """
-        On top of the requirements of the parent class,
-        the input dataframe should contain:
+        ???+ note "Additional construtor"
+            On top of the requirements of the parent class,
+            the input dataframe should contain:
 
-        (1) label_col_a and label_col_b for "label margins".
+            - label_col_a and label_col_b for "label margins".
+
+            | Param         | Type   | Description                  |
+            | :------------ | :----- | :--------------------------- |
+            | `df_dict`     | `dict` | `str` -> `DataFrame` mapping |
+            | `label_col_a` | `str`  | column for label set A       |
+            | `label_col_b` | `str`  | column for label set B       |
+            | `**kwargs`    |        | forwarded to `bokeh.plotting.figure` |
         """
         self.label_col_a = label_col_a
         self.label_col_b = label_col_b
         super().__init__(df_dict, **kwargs)
 
     def _setup_dfs(self, df_dict, **kwargs):
-        """Extending from the parent method."""
+        """
+        ???+ note "On top of the parent method, add column checks."
+            | Param      | Type   | Description                  |
+            | :--------- | :----- | :--------------------------- |
+            | `df_dict`  | `dict` | `str` -> `DataFrame` mapping |
+            | `**kwargs` |        | forwarded to the parent method |
+        """
         super()._setup_dfs(df_dict, **kwargs)
 
         for _key, _df in self.dfs.items():
@@ -282,7 +357,11 @@ class BokehMarginExplorer(BokehBaseExplorer):
 
     def plot(self, label, **kwargs):
         """
-        Plot the margins about a single label.
+        ???+ note "Plot the margins about a single label."
+            | Param      | Type   | Description                  |
+            | :--------- | :----- | :--------------------------- |
+            | `label`    |        | the label to plot about      |
+            | `**kwargs` |        | forwarded to plotting markers |
         """
 
         for _key, _source in self.sources.items():
@@ -321,16 +400,16 @@ class BokehMarginExplorer(BokehBaseExplorer):
 
 class BokehSnorkelExplorer(BokehBaseExplorer):
     """
-    Plot data points along with labeling function (LF) outputs.
+    ???+ note "Plot data points along with labeling function (LF) outputs."
 
-    Features:
+        Features:
 
-    - each labeling function corresponds to its own line_color.
-    - uses a different marker for each type of predictions: square for 'correct', x for 'incorrect', cross for 'missed', circle for 'hit'.
-      - 'correct': the LF made a correct prediction on a point in the 'labeled' set.
-      - 'incorrect': the LF made an incorrect prediction on a point in the 'labeled' set.
-      - 'missed': the LF is capable of predicting the target class, but did not make such prediction on the particular point.
-      - 'hit': the LF made a prediction on a point in the 'raw' set.
+        -   each labeling function corresponds to its own line_color.
+        -   uses a different marker for each type of predictions: square for 'correct', x for 'incorrect', cross for 'missed', circle for 'hit'.
+          -   'correct': the LF made a correct prediction on a point in the 'labeled' set.
+          -   'incorrect': the LF made an incorrect prediction on a point in the 'labeled' set.
+          -   'missed': the LF is capable of predicting the target class, but did not make such prediction on the particular point.
+          -   'hit': the LF made a prediction on a point in the 'raw' set.
     """
 
     SUBSET_GLYPH_KWARGS = {
@@ -349,10 +428,16 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
 
     def __init__(self, df_dict, **kwargs):
         """
-        On top of the requirements of the parent class,
-        the df_labeled input dataframe should contain:
+        ???+ note "Additional construtor"
+            Set up
 
-        (1) a "label" column for "ground truths".
+            -   a list to keep track of plotted labeling functions.
+            -   a palette for plotting labeling function predictions.
+
+            | Param       | Type   | Description                  |
+            | :---------- | :----- | :--------------------------- |
+            | `df_dict`   | `dict` | `str` -> `DataFrame` mapping |
+            | `**kwargs`  |        | forwarded to `bokeh.plotting.figure` |
         """
         super().__init__(df_dict, **kwargs)
 
@@ -362,9 +447,7 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
 
     def plot(self, *args, **kwargs):
         """
-        Overriding the parent method.
-
-        Plot only the raw subset.
+        ???+ note "Plot the raw subset in the background."
         """
         self.figure.circle(
             "x", "y", name="raw", source=self.sources["raw"], **self.glyph_kwargs["raw"]
@@ -375,12 +458,20 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
         self, lf, L_raw=None, L_labeled=None, include=("C", "I", "M"), **kwargs
     ):
         """
-        Plot a single labeling function.
+        ???+ note "Plot about a single labeling function."
+            | Param       | Type             | Description                  |
+            | :---------- | :--------------- | :--------------------------- |
+            | `lf`        | `callable`       | labeling function decorated by `@labeling_function()` from `hover.utils.snorkel_helper` |
+            | `L_raw`     | `np.ndarray`     | predictions, in decoded `str`, on the `"raw"` set |
+            | `L_labeled` | `np.ndarray`     | predictions, in decoded `str`, on the `"labeled"` set |
+            | `include`   | `tuple` of `str` | "C" for correct, "I" for incorrect, "M" for missed", "H" for hit: types of predictions to make visible in the plot |
+            | `**kwargs`  |                  | forwarded to plotting markers |
 
-        - param lf: labeling function decorated by @hover.utils.snorkel_helper.labeling_function()
-        - param L_raw: labeling function predictions, in decoded labels, on the raw df.
-        - param L_labeled: labeling function predictions, in decoded labels, on the labeled df.
-        - param include: subsets to show, which can be correct(C)/incorrect(I)/missed(M)/hit(H).
+
+            - lf: labeling function decorated by `@labeling_function()` from `hover.utils.snorkel_helper`
+            - L_raw: .
+            - L_labeled: .
+            - include: subsets to show, which can be correct(C)/incorrect(I)/missed(M)/hit(H).
         """
         # keep track of added LF
         self.lfs.append(lf)
@@ -454,7 +545,10 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
 
     def _view_correct(self, L_labeled):
         """
-        Determine the subset correctly labeled by a labeling function.
+        ???+ note "Determine the portion correctly labeled by a labeling function."
+            | Param       | Type         | Description                  |
+            | :---------- | :----------- | :--------------------------- |
+            | `L_labeled` | `np.ndarray` | predictions on the labeled subset |
         """
         agreed = self.dfs["labeled"]["label"].values == L_labeled
         attempted = L_labeled != module_config.ABSTAIN_DECODED
@@ -464,7 +558,10 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
 
     def _view_incorrect(self, L_labeled):
         """
-        Determine the subset incorrectly labeled by a labeling function.
+        ???+ note "Determine the portion incorrectly labeled by a labeling function."
+            | Param       | Type         | Description                  |
+            | :---------- | :----------- | :--------------------------- |
+            | `L_labeled` | `np.ndarray` | predictions on the labeled subset |
         """
         disagreed = self.dfs["labeled"]["label"].values != L_labeled
         attempted = L_labeled != module_config.ABSTAIN_DECODED
@@ -474,7 +571,11 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
 
     def _view_missed(self, L_labeled, targets):
         """
-        Determine the subset missed by a labeling function.
+        ???+ note "Determine the portion missed by a labeling function."
+            | Param       | Type          | Description                  |
+            | :---------- | :------------ | :--------------------------- |
+            | `L_labeled` | `np.ndarray`  | predictions on the labeled subset |
+            | `targets` | `list` of `str` | labels that the function aims for |
         """
         targetable = np.isin(self.dfs["labeled"]["label"], targets)
         abstained = L_labeled == module_config.ABSTAIN_DECODED
@@ -484,7 +585,10 @@ class BokehSnorkelExplorer(BokehBaseExplorer):
 
     def _view_hit(self, L_raw):
         """
-        Determine the subset hit by a labeling function.
+        ???+ note "Determine the portion hit by a labeling function."
+            | Param       | Type         | Description                  |
+            | :---------- | :----------- | :--------------------------- |
+            | `L_raw`     | `np.ndarray` | predictions on the raw subset |
         """
         indices = np.where(L_raw != module_config.ABSTAIN_DECODED)[0].tolist()
         view = CDSView(source=self.sources["raw"], filters=[IndexFilter(indices)])
