@@ -1,3 +1,8 @@
+"""
+???+ note "Neural network components."
+
+    `torch`-based template classes for implementing neural nets that work the most smoothly with `hover`
+"""
 import torch
 import torch.nn.functional as F
 from datetime import datetime
@@ -15,7 +20,7 @@ import numpy as np
 )
 def create_vector_net_from_module(specific_class, model_module_name, labels):
     """
-    Deprecated into a trivial invocation of VectorNet's class method.
+    ???+ warning "Deprecated into a trivial invocation of VectorNet's class method."
     """
     return specific_class.from_module(model_module_name, labels)
 
@@ -23,19 +28,24 @@ def create_vector_net_from_module(specific_class, model_module_name, labels):
 class VectorNet(Loggable):
 
     """
-    Simple transfer learning model: a user-supplied vectorizer followed by a neural net.
+    ???+ note "Simple transfer learning model: a user-supplied vectorizer followed by a neural net."
+        This is a parent class whose children may use different training schemes.
 
-    This is a parent class whose children may use different training schemes.
+        Coupled with:
 
-    Please refer to hover.utils.torch_helper.VectorDataset and vector_dataloader for more info.
+        -   `hover.utils.torch_helper.VectorDataset`
+        -   `hover.utils.torch_helper.vector_dataloader`
     """
 
     def __init__(self, vectorizer, architecture, state_dict_path, labels):
         """
-        - param vectorizer(callable): a function that converts any string to a NumPy 1-D array.
-        - param architecture(class): a `torch.nn.Module` child class to be instantiated into a neural net.
-        - param state_dict_path(str): path to a PyTorch state dict that matches the architecture.
-        - param labels(list of str): the classification labels, e.g. ["POSITIVE", "NEGATIVE"].
+        ???+ note "Create the `VectorNet`, loading parameters if available."
+            | Param             | Type       | Description                          |
+            | :---------------- | :--------- | :----------------------------------- |
+            | `vectorizer`      | `callable` | the feature -> vector function       |
+            | `architecture`    | `class`    | a `torch.nn.Module` child class      |
+            | `state_dict_path` | `str`      | path to a (could-be-empty) `torch` state dict |
+            | `labels`          | `list`     | list of `str` classification labels  |
         """
 
         # set up label conversion
@@ -74,10 +84,12 @@ class VectorNet(Loggable):
     @classmethod
     def from_module(cls, model_module, labels):
         """
-        Create a VectorNet model, or of its child class.
+        ???+ note "Create a VectorNet model from a loadable module."
 
-        - param model_module(module or str): (path to) a local Python module in the working directory whose __init__.py file contains a get_vectorizer() callable, get_architecture() callable, and a get_state_dict_path() callable.
-        - param labels(list of str): the classification labels, e.g. ["POSITIVE", "NEGATIVE"].
+            | Param          | Type       | Description                          |
+            | :------------- | :--------- | :----------------------------------- |
+            | `model_module` | `module` or `str` | (path to) a local Python workspace module which contains a get_vectorizer() callable, get_architecture() callable, and a get_state_dict_path() callable |
+            | `labels`       | `list`     | list of `str` classification labels  |
         """
         if isinstance(model_module, str):
             from importlib import import_module
@@ -96,7 +108,10 @@ class VectorNet(Loggable):
 
     def save(self, save_path=None):
         """
-        Save the current state dict with authorization to overwrite.
+        ???+ note "Save the current state dict with authorization to overwrite."
+            | Param       | Type  | Description                           |
+            | :---------- | :---- | :------------------------------------ |
+            | `save_path` | `str` | option alternative path to state dict |
         """
         if save_path is None:
             save_path = self.nn_update_path
@@ -104,19 +119,22 @@ class VectorNet(Loggable):
 
     def adjust_optimizer_params(self):
         """
-        Dynamically change parameters of the neural net optimizer.
+        ???+ note "Dynamically change parameters of the neural net optimizer."
 
-        - Intended to be polymorphic in child classes and to be called per epoch.
+            - Intended to be polymorphic in child classes and to be called per epoch.
         """
         for _group in self.nn_optimizer.param_groups:
             _group.update(self._dynamic_params["optimizer"])
 
     def predict_proba(self, inps):
         """
-        End-to-end single/multi-piece prediction from inp to class probabilities.
+        ???+ note "End-to-end single/multi-piece prediction from inp to class probabilities."
+            | Param  | Type    | Description                          |
+            | :----- | :------ | :----------------------------------- |
+            | `inps` | dynamic | (a list of) input features to vectorize |
         """
         # if the input is a single piece of inp, cast it to a list
-        FLAG_SINGLE = isinstance(inps, str)
+        FLAG_SINGLE = not isinstance(inps, list)
         if FLAG_SINGLE:
             inps = [inps]
 
@@ -134,15 +152,20 @@ class VectorNet(Loggable):
 
     def manifold_trajectory(self, inps, method="umap", **kwargs):
         """
-        TODO: need a clean way to pass kwargs to dimensionality reduction.
+        ???+ note "**UPCOMING**"
+            Compute a trajectory of manifold propagation through the neural net.
 
-        1. vectorize inps
-        2. forward propagate, keeping intermediates
-        3. fit intermediates to 2D manifolds
-        4. fit manifolds using Procrustes shape analysis
-        5. fit shapes to trajectory splines
+            1. vectorize inps
+            2. forward propagate, keeping intermediates
+            3. fit intermediates to 2D manifolds
+            4. fit manifolds using Procrustes shape analysis
+            5. fit shapes to trajectory splines
 
-        - param inps(list): input to calculate the manifold profile from.
+            | Param    | Type    | Description                          |
+            | :------- | :------ | :----------------------------------- |
+            | `inps`   | dynamic | (a list of) input features to vectorize |
+            | `method` | `str`   | reduction method: `"umap"` or `"ivis"`  |
+            | `**kwargs` | | kwargs to forward to dimensionality reduction |
         """
         from hover.core.representation.manifold import LayerwiseManifold
         from hover.core.representation.trajectory import manifold_spline
@@ -166,7 +189,12 @@ class VectorNet(Loggable):
 
     def evaluate(self, dev_loader, verbose=1):
         """
-        Evaluate the neural network against a dev set.
+        ???+ note "Evaluate the VecNet against a dev set."
+
+            | Param        | Type         | Description                |
+            | :----------- | :----------- | :------------------------- |
+            | `dev_loader` | `torch.utils.data.DataLoader` | dev set   |
+            | `verbose`    | `int`        | verbosity for logging      |
         """
         self.nn.eval()
         true = []
@@ -198,10 +226,17 @@ class VectorNet(Loggable):
 
     def train(self, train_loader, dev_loader=None, epochs=1, verbose=1):
         """
-        Train the neural network.
+        ???+ note "Train the neural network part of the VecNet."
 
-        - This method is a vanilla template and is intended to be overridden in child classes.
-        - Also intended to be coupled with self.train_batch().
+            - This method is a vanilla template and is intended to be overridden in child classes.
+            - Also intended to be coupled with self.train_batch().
+
+            | Param          | Type         | Description                |
+            | :------------- | :----------- | :------------------------- |
+            | `train_loader` | `torch.utils.data.DataLoader` | train set |
+            | `dev_loader`   | `torch.utils.data.DataLoader` | dev set   |
+            | `epochs`       | `int`        | number of epochs to train  |
+            | `verbose`      | `int`        | verbosity for logging      |
         """
         train_info = []
         for epoch_idx in range(epochs):
@@ -214,9 +249,17 @@ class VectorNet(Loggable):
 
     def train_epoch(self, train_loader, *args, **kwargs):
         """
-        Train the neural network for one epoch.
+        ???+ note "Train the neural network for one epoch."
 
-        - Supports flexible args and kwargs for child classes that may implement self.train() and self.train_batch() differently.
+            - Supports flexible args and kwargs for child classes that may implement self.train() and self.train_batch() differently.
+
+            | Param          | Type         | Description                |
+            | :------------- | :----------- | :------------------------- |
+            | `train_loader` | `torch.utils.data.DataLoader` | train set |
+            | `dev_loader`   | `torch.utils.data.DataLoader` | dev set   |
+            | `verbose`      | `int`        | verbosity for logging      |
+            | `*args`        | | arguments to forward to `train_batch`   |
+            | `**kwargs`     | | kwargs to forward to `train_batch`      |
         """
         self.adjust_optimizer_params()
         for batch_idx, (loaded_input, loaded_output, _) in enumerate(train_loader):
@@ -225,7 +268,13 @@ class VectorNet(Loggable):
 
     def train_batch(self, loaded_input, loaded_output, verbose=1):
         """
-        Train the neural network for one batch.
+        ???+ note "Train the neural network for one batch."
+
+            | Param           | Type           | Description           |
+            | :-------------- | :------------- | :-------------------- |
+            | `loaded_input`  | `torch.Tensor` | input tensor          |
+            | `loaded_output` | `torch.Tensor` | output tensor         |
+            | `verbose`       | `int`          | verbosity for logging |
         """
         self.nn.train()
         input_tensor = loaded_input.float()
